/*global angular */

/*
	Copyright 2013 wrongcoder@gmail.com
	Ludum Dare 26 compo entry
	No offense is intended to any real-life movements or organisations
	Have a potato day.
*/

(function () {
	"use strict";

	var game = angular.module("game", []);

	game.controller("controller", ['$scope', '$timeout', function ($scope, $timeout) {

		// Game state
		$scope.day = 1;
		$scope.time = 0;
		$scope.sales = 0;
		$scope.complaints = 0;
		$scope.protests = 0;
		$scope.wastedPotatoes = 0;
		$scope.busyTurns = 0;
		$scope.busyDone = nothing;
		$scope.cookedFries = 0;
		$scope.cash = 250;
		$scope.fixedCosts = 18;
		$scope.cookTime = 4;
		$scope.servingsPerPound = 5;
		$scope.potato = 0;
		$scope.potatoPrice = 0.60;
		$scope.tofu = 0;
		$scope.tofuPrice = 4.25;
		$scope.oilChanged = 0;
		$scope.oilPrice = 30;
		$scope.reputation = 0;
		$scope.priceToleranceFactor = 1;
		resetQueue();

		$scope.totalSales = 0;
		$scope.totalComplaints = 0;
		$scope.totalProtests = 0;
		$scope.totalPotatoes = 0;
		$scope.totalWastedFries = 0;
		$scope.totalWastedPotatoes = 0;

		// Views
		$scope.showIntro = true;
		$scope.showNews = false;
		$scope.showNewsOk = true;
		$scope.showSetup = false;
		$scope.showSim = false;
		$scope.showResult = false;
		$scope.showLose = false;
		$scope.showVandalism = false;

		// View data
		$scope.title = 'Ethical Spud';
		$scope.price = 2;
		$scope.batchSize = 2;
		$scope.batchAge = 12;
		$scope.potatoPurchase = 10;
		$scope.tofuPurchase = 0;
		$scope.oilPurchase = 0;
		$scope.activity = [];
		$scope.bodyClass = '';

		// Static data
		var tofuDay = 7;
		var endDay = 31;
		var news = [];
		var events = [];
		var inflation = Math.pow(Math.E, Math.log(1.60) / 365);

		// --------------------------------------------------
		// Controller actions

		$scope.dismissIntro = function () {
			$scope.showIntro = false;
			$scope.showSetup = true;
		};

		$scope.dismissNews = function () {
			if ($scope.day != endDay) {
				$scope.showNews = false;
				$scope.showSetup = true;
			} else {
				$scope.showNewsOk = false;
				$scope.showWin = true;
			}
		};

		$scope.finishSetup = function () {
			$scope.showSetup = false;
			$scope.showSim = true;
			startSimulation();
		};

		$scope.finishDay = function () {
			$scope.day += 1;
			$scope.time = 0;
			$scope.showSim = false;
			$scope.showResult = false;

			if (news[$scope.day] !== undefined) {
				$scope.showNews = true;
			} else {
				$scope.showSetup = true;
			}
		};

		$scope.continueAfterWin = function () {
			$scope.showWin = false;
			$scope.showNewsOk = true;
			$scope.showNews = false;
			$scope.showSetup = true;
		};

		// --------------------------------------------------
		// Simulation

		function startSimulation() {
			$scope.sales = 0;
			$scope.complaints = 0;
			$scope.protests = 0;
			$scope.wastedPotatoes = 0;
			$scope.cookedFries = 0;
			$scope.busyTurns = 0;
			$scope.busyDone = nothing;
			$scope.activity = [];
			resetQueue();

			$scope.cash -= $scope.fixedCosts;
			$scope.potato += $scope.potatoPurchase;
			$scope.cash -= $scope.potatoPurchase * $scope.potatoPrice;
			$scope.tofu += $scope.tofuPurchase;
			$scope.cash -= $scope.tofuPurchase * $scope.tofuPrice;

			if ($scope.oilPurchase) {
				$scope.cash -= $scope.oilPrice;
				$scope.oilChanged = $scope.day;
			}
			$scope.oilPurchase = 0;

			if (totalPounds() > 0) {
				cookFries();
			} else {
				logActivity("You wonder what you're going to do all day with no food in stock.");
			}

			$scope.bodyClass = 'wait';
			$timeout(tickSimulation, delayTime());
		}

		function tickSimulation() {
			filterImpatientCustomers();

			$scope.time += 1;

			if ($scope.time >= 108) {
				$timeout(endSimulation, 0);
				return;
			}

			if (isCustomerQueueing()) {
				customerApproaches();
			}
			if ($scope.time >= 20 && $scope.time <= 52 && isCustomerQueueing()) {
				customerApproaches();
			}
			if ($scope.time >= 24 && $scope.time <= 48 && isCustomerQueueing()) {
				customerApproaches();
			}
			if ($scope.time >= 39 && $scope.time <= 44 && isCustomerQueueing()) {
				customerApproaches();
			}

			var timeId = $scope.day + '-' + $scope.time;
			if (events[timeId] !== undefined) {
				var continueCallback = processTurn;
				var stopCallback = endSimulation;
				events[timeId](continueCallback, stopCallback);
				return;
			}

			processTurn();
		}

		function processTurn() {
			if (getFriesAge() > $scope.batchAge) {
				var wastedPotatoes = $scope.cookedFries / $scope.servingsPerPound * (1 - tofuContent());
				$scope.wastedPotatoes += wastedPotatoes;
				$scope.totalWastedPotatoes += wastedPotatoes;
				$scope.totalWastedFries += $scope.cookedFries;
				logActivity("You throw out " + $scope.cookedFries + " servings of stale fries.");
				$scope.cookedFries = 0;
			}

			if ($scope.busyTurns > 1) {
				$scope.busyTurns -= 1;
			} else if ($scope.busyTurns == 1) {
				$scope.busyDone();
				$scope.busyDone = nothing;
				$scope.busyTurns = 0;
			} else if ($scope.cookedFries <= 0 && totalPounds() > 0 && $scope.time < 108 - ($scope.cookTime + $scope.batchSize * $scope.servingsPerPound / 10)) {
				cookFries();
			} else if (lengthQueue() > 0) {
				processCustomer();
			}

			$timeout(tickSimulation, delayTime());
		}

		function endSimulation() {
			$scope.bodyClass = '';
			$scope.time = Math.max(108, $scope.time);
			if ($scope.cash > 0) {
				logActivity("You close up shop for the day and drive home.");
				$scope.showResult = true;
			} else {
				logActivity("You lock up your beloved food truck for the last time and walk away, defeated.");
				$scope.showLose = true;
			}

			$scope.potatoPrice *= inflation;
			$scope.tofuPrice *= inflation;
			$scope.oilPrice *= inflation;
			$scope.fixedCosts *= inflation;
			$scope.priceToleranceFactor *= 1 + (inflation - 1) / 2;
		}

		function delayTime() {
			return 240;
		}

		function totalPounds() {
			return $scope.potato + $scope.tofu;
		}

		function tofuContent() {
			return totalPounds() == 0 ? 0 : $scope.tofu / totalPounds();
		}

		function cookFries() {
			var batch = Math.min($scope.batchSize, totalPounds());
			var tofuUsed = batch * tofuContent();
			var potatoUsed = batch - tofuUsed;
			var friesCreated = $scope.servingsPerPound * batch;

			$scope.tofu -= tofuUsed;
			$scope.potato -= potatoUsed;
			$scope.totalPotatoes += potatoUsed;
			logActivity("You start cooking a batch of fries.");

			$scope.busyTurns = $scope.cookTime;
			$scope.busyDone = function () {
				$scope.cookedFries += friesCreated;
				$scope.friesCookedTime = $scope.time;
				logActivity("You finish cooking a batch of fries.");
			};
		}

		function sellFries() {
			$scope.sales += 1;
			$scope.totalSales += 1;
			$scope.cookedFries -= 1;
			$scope.cash += $scope.price;
			$scope.lastServed = $scope.time;
			logActivity("You sold some fries!");
		}

		function isCustomerQueueing() {
			var reputationCap = 200;
			var maxProbability = 0.80;
			var minProbability = 0.04;
			var probability = $scope.reputation * (maxProbability - minProbability) / reputationCap + minProbability;
			probability = Math.max(probability, minProbability);
			probability = Math.min(probability, maxProbability);

			var chance = Math.random();
			return chance < probability;
		}

		function customerApproaches() {
			var customer = {
				arrived: $scope.time,
				waitTolerance: Math.max(1, 5.5 + normalRandom(3.5)),
				priceTolerance: Math.max(0, 2.75 * $scope.priceToleranceFactor + normalRandom(2.75)),
				tofuTolerance: Math.max(0, 0.65 + normalRandom(0.75)),
				potatoTolerance: Math.max(0, (isTofuExists() ? 0.85 : 1.10) + normalRandom(0.65)),
				oilTolerance: Math.max(0, 10 + normalRandom(4)),
				staleTolerance: Math.max(2, 15 + normalRandom(9))
			};

			if ($scope.price > customer.priceTolerance) {
				$scope.reputation -= 0.25;

				var rand = Math.random();
				if (rand < 0.25) {
					logActivity("Customer: \"That's too expensive!\"");
				} else if (rand < 0.50) {
					logActivity("Customer: \"Your prices are outrageous!\"");
				} else if (rand < 0.75) {
					logActivity("Customer: \"This inflation is out of control!\"");
				} else {
					logActivity("Customer: \"I can't afford that!\"");
				}
				return;
			}

			pushQueue(customer);
		}

		function filterImpatientCustomers() {
			var remaining = [];

			if (lengthQueue() <= 0) {
				return;
			}

			while (lengthQueue() > 0) {
				var customer = popQueue();
				var rand = Math.random();
				if ($scope.time - customer.arrived > customer.waitTolerance) {
					$scope.reputation -= 0.5;
					if (rand < 0.3) {
						logActivity("Customer: \"This is taking too long. I'm leaving!\"");
					} else if (rand < 0.6) {
						logActivity("Customer: \"I don't have all day to wait around for fries!\"");
					} else if (rand < 0.9) {
						logActivity("Customer: \"Come on! You're selling fries, not a three course meal!\"");
					} else {
						logActivity("Customer: \"You're too slow! Why don't you hire more people?\"");
					}
				} else {
					remaining.push(customer);
				}
			}

			resetQueue();
			for (var ix = 0; ix < remaining.length; ix++) {
				pushQueue(customer);
			}
		}

		function processCustomer() {
			var rand = Math.random();
			var customer = popQueue();

			if ($scope.cookedFries <= 0) {
				$scope.reputation -= 0.25;

				if (rand < 0.5) {
					logActivity("Customer: \"What do you mean, you're not making any more fries?\"");
				} else {
					logActivity("Customer: \"So are you selling fries or aren't you?\"");
				}
				return;
			}

			sellFries();

			if (tofuContent() > customer.tofuTolerance) {
				$scope.reputation -= 3;
				incrementComplaints();
				if (rand < 0.333) {
					logActivity("Customer: \"This tastes like tofu! Ew! Yuck!\"");
				} else if (rand < 0.666) {
					logActivity("Customer: \"Did you drop some tofu in here by mistake?\"");
				} else {
					logActivity("Customer: \"Tofu! So you're bowing down to the pressure, huh?\"");
				}
			} else if (getOilAge() > customer.oilTolerance) {
				$scope.reputation -= 3;
				incrementComplaints();
				if (rand < 0.5) {
					logActivity("Customer: \"This stinks! Do you ever change your fryer oil?\"");
				} else {
					logActivity("Customer: \"I think these fries have gone rancid!\"");
				}
			} else if (getFriesAge() > customer.staleTolerance) {
				$scope.reputation -= 3;
				incrementComplaints();
				if (rand < 0.333) {
					logActivity("Customer: \"These fries are stale!\"");
				} else if (rand < 0.666) {
					logActivity("Customer: \"You serve cold fries? That's disgusting!\"");
				} else {
					logActivity("Customer: \"I wouldn't serve these soggy fries to my dog!\"");
				}
			} else if (1 - tofuContent() > customer.potatoTolerance) {
				$scope.reputation -= 2;
				incrementProtests();
				if (rand < 0.333) {
					logActivity("Customer: \"I can't believe you're killing potatoes!\"");
				} else if (rand < 0.666) {
					logActivity("Customer: \"You cut up these poor innocent potatoes just to make fries?\"");
				} else {
					logActivity("Customer: \"That looks like potato flesh! Someone should shut you down!\"");
				}
			} else {
				$scope.reputation += 2;
			}
		}

		// --------------------------------------------------
		// Helpers

		function nothing() {
		}

		function isTofuExists() {
			return $scope.day >= tofuDay;
		}

		function getOilAge() {
			return $scope.day - $scope.oilChanged;
		}

		function getFriesAge() {
			if ($scope.cookedFries > 0) {
				return $scope.time - $scope.friesCookedTime;
			} else {
				return 0;
			}
		}

		function timeAsString() {
			var hour = 9 + Math.floor($scope.time / 12);
			var minute = 5 * ($scope.time % 12);
			return (hour < 10 ? ' ' : '') + hour + ":" + (minute < 10 ? '0' : '') + minute;
		}

		function getCurrentState() {
			if ($scope.time == $scope.lastServed) {
				return "serving";
			}
			if ($scope.busyTurns > 0) {
				return "busy";
			}
			return "idle";
		}

		function logActivity(message) {
			$scope.activity.push(timeAsString() + " " + message);
		}

		function normalRandom(stdDev) {
			var count = 10;
			var value = 0;
			for (var i = 0; i < count; i++) {
				value += (Math.random() - 0.5) * 5 * stdDev;
			}
			return value / count;
		}

		function incrementComplaints() {
			$scope.complaints += 1;
			$scope.totalComplaints += 1;
		}

		function incrementProtests() {
			$scope.protests += 1;
			$scope.totalProtests += 1;
		}

		$scope.getTime = timeAsString;
		$scope.getOilAge = getOilAge;
		$scope.getCurrentState = getCurrentState;
		$scope.isTofuExists = isTofuExists;

		$scope.getNews = function () {
			if (news[$scope.day] !== undefined) {
				return news[$scope.day];
			} else {
				return "Nothing happened. How boring.";
			}
		};

		function resetQueue() {
			$scope.queue = { elt: [], next: 0 };
		}

		function pushQueue(customer) {
			$scope.queue.elt[$scope.queue.elt.length] = customer;
		}

		function popQueue() {
			return $scope.queue.elt[$scope.queue.next++];
		}

		function lengthQueue() {
			return $scope.queue.elt.length - $scope.queue.next;
		}

		$scope.lengthQueue = lengthQueue;

		// --------------------------------------------------
		// Tables

		news[2] = "Protesters spotted picketing fast food restaurants";
		news[6] = "City on Graffiti Alert! No graffiti permitted after 5pm";
		news[tofuDay] = "Scientists invent PotaTofu, a potato substitute that tastes just like the real thing";
		news[tofuDay+3] = "Ethical Potatologist disputes findings that PotaTofu may cause bowel irritation";
		news[tofuDay+6] = "Economists blame overseas investors for record-high inflation";
		news[endDay-12] = "School for Potato Studies lobbies for all potato-harming business licenses to be revoked";
		news[endDay-8] = "Potato-eaters beware: City-wide potato awareness campaign starts next week";
		news[endDay-3] = "Mayor under pressure to reduce potato harm";
		news[endDay-1] = "Mayor in talks with Potato Studies lobbyists";
		news[endDay] = "Mayor: All potato-harming business licenses revoked immediately";

		if (endDay - tofuDay <= 19) {
			window.alert("Internal error: tofu/end too close");
		}

		events['2-28'] = function (continueCallback, stopCallback) {
			incrementProtests();
			logActivity("You hear someone yell, \"Fries are murder!\"");
			continueCallback();
		};

		events['3-48'] = function (continueCallback, stopCallback) {
			incrementProtests();
			$scope.showVandalism = true;
			logActivity("You hear someone yell, \"Potatoes are people too!\"");
			continueCallback();
		};

		events['3-97'] = function (continueCallback, stopCallback) {
			incrementProtests();
			logActivity("You hear someone yell, \"You wouldn't eat a poor defenseless daffodil! How could you eat a potato?\"");
			continueCallback();
		};

		events['6-92'] = function (continueCallback, stopCallback) {
			$scope.showVandalism = false;
			logActivity("You spend the rest of the day cleaning your signage.");
			stopCallback();
		};

		events['7-31'] = function (continueCallback, stopCallback) {
			incrementProtests();
			logActivity("You hear someone yell, \"Starch mouth! You will burn for this!\"");
			continueCallback();
		};

		events['7-74'] = function (continueCallback, stopCallback) {
			incrementProtests();
			logActivity("You hear someone yell, \"Your dog doesn't need to eat potatoes to be healthy! You don't either!\"");
			continueCallback();
		};

		events[tofuDay + '-68'] = function (continueCallback, stopCallback) {
			logActivity("You hear someone yell, \"PotaTofu is the devil's food!\"");
			continueCallback();
		};

		events[(tofuDay + 1) + '-33'] = function (continueCallback, stopCallback) {
			if (tofuContent() < 0.75) {
				incrementProtests();
				logActivity("You hear someone yell, \"Save the potato!\"");
			}
			continueCallback();
		};

		events[(tofuDay + 2) + '-33'] = function (continueCallback, stopCallback) {
			if (tofuContent() < 0.75) {
				incrementProtests();
				logActivity("You hear someone yell, \"Save the potato!\"");
			}
			continueCallback();
		};

		events[(tofuDay + 3) + '-33'] = function (continueCallback, stopCallback) {
			if (tofuContent() < 0.75) {
				incrementProtests();
				logActivity("You hear someone yell, \"Save the potato!\" Again?");
			}
			continueCallback();
		};

		events[(tofuDay + 3) + '-33'] = function (continueCallback, stopCallback) {
			if (tofuContent() < 0.75) {
				incrementProtests();
				logActivity("You find a note left on your counter. It reads: \"We tried peaceful protest but you continue to harm the potato. Let this be a warning to you, murderer.\"");
			}
			continueCallback();
		};
		events[(tofuDay + 3) + '-37'] = function (continueCallback, stopCallback) {
			if (tofuContent() < 0.75) {
				logActivity("You feel the ground sway beneath you!");
				logActivity("You hop down and have a look at your truck. Your tires have been slashed.");
				logActivity("You spend the rest of the day talking to the police and having your tired replaced.");
				stopCallback();
			} else {
				continueCallback();
			}
		};

		$scope.priceOptions = (function () {
			var options = [];
			for (var i = 0; i < 15; i++) {
				options.push(i);
				options.push(i + 0.25);
				options.push(i + 0.50);
				options.push(i + 0.75);
			}
			return options;
		}());

		$scope.ageOptions = (function () {
			var options = [];
			for (var i = 0; i <= 72; i++) {
				options.push({label: i * 5, value: i});
			}
			return options;
		}());

	}]);

	game.directive('scrolledAppenderElement', function () {
		return function (scope, element, attributes) {
			var domParentElement = element.parent()[0];
			domParentElement.scrollTop = domParentElement.scrollHeight;
		}
	});

}());
